/**
 * Copyright (c) 2019 Eclipse contributors and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v2.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v20.html
 */
package org.eclipse.oomph.mirror.index;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class MirrorIndexGenerator {

	private static String FILES = "%FILES%";

	private static String ROOT_ECLIPSE = "eclipse";

	private static Pattern MIRROR_PATTERN = Pattern
			.compile("<mirror url=\"([^\"]+?)/?/favicon.ico\" label=\"\\[(.+)\\] ([^(]+).*\"");

	private File outputFolder;

	private String summaryLocation;

	private String mirrorPhpLocation;

	private Map<String, String> sizes = new LinkedHashMap<String, String>();

	private Map<String, String> dates = new LinkedHashMap<String, String>();

	private Map<String, String> times = new LinkedHashMap<String, String>();

	private Map<String, String> fileCounts = new LinkedHashMap<String, String>();

	private Map<String, List<String>> children = new LinkedHashMap<String, List<String>>();

	private float totalSize;

	public MirrorIndexGenerator(String summaryLocation, String mirrorPhpLocation) {
		this.summaryLocation = summaryLocation;
		this.mirrorPhpLocation = mirrorPhpLocation;
		outputFolder = new File(summaryLocation).getAbsoluteFile().getParentFile();
	}

	public static void main(String[] args) throws IOException, InterruptedException {
		new MirrorIndexGenerator(args[0], args[1]).run();
	}

	private void run() throws IOException, InterruptedException {

		updateMirrors();

		List<String> lines = MirrorIndex.readLines(new File(summaryLocation));
		for (String line : lines) {
			String[] parts = line.split("\\s");
			String size = parts[0];
			String location = parts[1].length() == 1 ? "eclipse" : "eclipse" + parts[1];
			location = location.replace("/./", "");
			String date = parts[2];
			String time = parts[3];
			String fileCount = parts[4];
			String fileSize = parts[5];
			handle(location, size, date, time, fileCount, fileSize);
		}

		Comparator<String> sizeComparator = new Comparator<String>() {
			@Override
			public int compare(String o1, String o2) {
				String size1 = sizes.get(o1);
				String size2 = sizes.get(o2);
				return Float.compare(MirrorIndex.parseSize(size2), MirrorIndex.parseSize(size1));
			}
		};

		for (List<String> list : children.values()) {
			Collections.sort(list, sizeComparator);
		}

		List<String> children = this.children.get("eclipse");

		totalSize = MirrorIndex.parseSize(sizes.get("eclipse"));

		visit("eclipse", children, null, null);
	}

	private static class Mirror implements Comparable<Mirror> {
		public String uri;
		public String location;
		public String label;

		public Mirror(String uri, String location, String label) {
			this.uri = uri;
			this.location = location;
			this.label = label;
		}

		@Override
		public int compareTo(Mirror mirror) {
			int result = location.compareToIgnoreCase(mirror.location);
			if (result == 0) {
				result = label.compareToIgnoreCase(mirror.label);
				if (result == 0) {
					uri.compareToIgnoreCase(mirror.uri);
				}
			}

			return result;
		}

		@Override
		public String toString() {
			return "Mirror [location=" + location + ", label=" + label + ", uri=" + uri + "]";
		}
	}

	private void updateMirrors() throws IOException, InterruptedException {
		List<Mirror> mirrors = getMirrors();
		List<String> lines = MirrorIndex.readLines(new File(mirrorPhpLocation));
		if (!mirrors.isEmpty()) {
			int begin = lines.indexOf("### MIRRORS-BEGIN ###");
			int end = lines.indexOf("### MIRRORS-END ###");
			int index = begin + 1;
			for (int i = index; i < end; ++i) {
				lines.remove(index);
			}

			for (Mirror mirror : mirrors) {
				lines.add(index++, "\"" + mirror.uri + "\", \"" + mirror.location + "\", \"" + mirror.label + "\",");
			}
		}

		File mirrorPhpOut = new File(outputFolder, "mirror.php");
		try (PrintStream out = new PrintStream(mirrorPhpOut, "UTF-8")) {
			for (String line : lines) {
				out.println(line);
			}
		}
	}

	private List<Mirror> getMirrors() throws IOException, InterruptedException {
		List<String> lines = MirrorIndex
				.readLines("http://www.eclipse.org/downloads/download.php?file=/favicon.ico&format=xml");
		List<Mirror> mirrors = new ArrayList<Mirror>();
		for (String mirror : lines) {
			System.out.println(mirror);
			Matcher matcher = MIRROR_PATTERN.matcher(mirror);
			if (matcher.find()) {
				mirrors.add(new Mirror(matcher.group(1), matcher.group(2), matcher.group(3).trim()));
			}
		}

		if (mirrors.isEmpty()) {
			System.out.println("No mirrors found; hard-coded defaults will be used.");
		} else if (mirrors.isEmpty()) {
			mirrors.add(new Mirror("https://download.eclipse.org", "Canada", "<b>Main eclipse.org download area</b>"));

			Collections.sort(mirrors);

			for (Mirror mirror : mirrors) {
				System.out.println(mirror);
			}
		}

		return mirrors;
	}

	private static String formatPercent(float numerator, float denominator) {
		if (denominator == 0f) {
			return "100.00%";
		} else {
			float value = 100f * numerator / denominator;
			NumberFormat instance = NumberFormat.getInstance(Locale.US);
			instance.setMaximumFractionDigits(2);
			instance.setMinimumFractionDigits(2);
			return instance.format(value) + '%';
		}
	}

	private void visit(String location, List<String> children, PrintStream out, String parentMirrorLinkPrefix)
			throws IOException {
		String[] segments = location.split("/", 0);
		String name = segments[segments.length - 1];
		boolean isFiles = FILES.equals(name);
		String size = sizes.get(location);

		// Print this folder's index page.
		if (!isFiles) {
			File file = new File(outputFolder, cleanse(location));
			if (!file.mkdir() && !file.isDirectory()) {
				System.err.println("###" + file);
			}

			File index = new File(file, "index.html");
			PrintStream indexOut = new PrintStream(index, "UTF-8");
			String mirrorLinkPrefix = printHTMLPrefix(indexOut, location);

			if (out != null) {
				// Print the parent entry.
				indexOut.println(//
						"  <tr><td style='padding-right: 2em;'><a href='../index.html#" + cleanse(name)
								+ "'>..</a></td>" + //
								"<td style='text-align: right; padding-right: 2em;'>" + size + "</td>" + //
								"<td style='text-align: right; padding-right: 2em;'></td>" + //
								"<td style=''>" + dates.get(location) + "</td></tr>");
			}

			for (String child : children) {
				List<String> grandChildren = this.children.get(child);
				visit(child, grandChildren, indexOut, mirrorLinkPrefix);
			}

			printHTMLSuffix(indexOut);

			indexOut.close();
		}

		// Print into the parent folder's table of contents.
		if (out != null) {
			String parentLocation = location.substring(0, location.lastIndexOf('/'));
			String parentSize = sizes.get(parentLocation);
			float parentSizeValue = MirrorIndex.parseSize(parentSize);
			float sizeValue = MirrorIndex.parseSize(size);
			String formattedPercent = formatPercent(sizeValue, parentSizeValue);
			String specializedSize = size.endsWith("G") ? "<b>" + size + "</b>" : size;

			String label = isFiles
					? parentMirrorLinkPrefix + "<em>Local Files: " + fileCounts.get(location) + "</em></a>"
					: "<a id='" + cleanse(name) + "' href='" + cleanse(name) + "/index.html'>" + name + "</a>";

			out.println(//
					"  <tr><td style='padding-right: 2em;'>" + label + "</td>" + //
							"<td style='text-align: right; padding-right: 2em;'>" + specializedSize + "</td>" + //
							"<td style='text-align: right; padding-right: 2em;'>" + formattedPercent + "</td>" + //
							"<td style=''>" + dates.get(location) + "</td></tr>");
		}
	}

	private String cleanse(String location) {
		return location.replace(':', '_');
	}

	private String printHTMLPrefix(PrintStream out, String location) {
		class BreadCrumb {
			public String name;
			public String size;
			public String percent;
			public String eclipseRelativePath;
			public String navigationHref = "";
			public String rootHref = "..";

			public BreadCrumb(String name, String size) {
				this.name = name;
				this.size = size;
				if (name == ROOT_ECLIPSE) {
					this.percent = "";
					this.eclipseRelativePath = "";
				} else {
					this.percent = formatPercent(MirrorIndex.parseSize(size), totalSize);
					this.eclipseRelativePath = name;
				}
			}

			public void addTo(List<BreadCrumb> breadCrumbs) {
				for (int i = 0, size = breadCrumbs.size(); i < size; i++) {
					BreadCrumb breadCrumb = breadCrumbs.get(i);
					if (i == 0) {
						navigationHref = breadCrumb.rootHref + "/index.html#" + cleanse(breadCrumb.name);
						rootHref = breadCrumb.rootHref = "../" + breadCrumb.rootHref;
					} else {
						breadCrumb.rootHref = "../" + breadCrumb.rootHref;
					}

					if (name != ROOT_ECLIPSE) {
						breadCrumb.eclipseRelativePath = name + "/" + breadCrumb.eclipseRelativePath;
					}
				}

				breadCrumbs.add(0, this);
			}

			@Override
			public String toString() {
				return "BreadCrumb [name=" + name + ", size=" + size + ", percent=" + percent + ", eclipseRelativePath="
						+ eclipseRelativePath + ", navigationHref=" + navigationHref + ", rootHref=" + rootHref + "]";
			}

			public String getMirorrLinkPrefix() {
				return "<a href='" + rootHref + "/mirror.php?location=" + eclipseRelativePath
						+ "' target='eclipse_mirror'>";
			}

			public String toHTML(boolean isMirrorLink) {
				return getMirorrLinkPrefix() + "<img src='" + rootHref + "/Nav.png'/></a> "
						+ (isMirrorLink ? getMirorrLinkPrefix() + name + "</a>"
								: "<a href='" + navigationHref + "'>" + name + "</a>") //
						+ " <small><em style='color: DarkOrange;'>" + size + "</em></small>" //
						+ " <small>"
						+ (eclipseRelativePath.isEmpty() ? "" : (percent.equals("0.00%") ? "&lt; 0.01%" : "" + percent))
						+ "</small>";
			}
		}

		String baseLocation = location;
		List<BreadCrumb> breadCrumbs = new ArrayList<BreadCrumb>();
		for (int index = baseLocation.lastIndexOf('/'); index != -1; index = baseLocation.lastIndexOf('/')) {
			String baseSize = sizes.get(baseLocation);
			String childName = baseLocation.substring(index + 1);

			new BreadCrumb(childName, baseSize).addTo(breadCrumbs);

			baseLocation = baseLocation.substring(0, index);
		}

		new BreadCrumb(ROOT_ECLIPSE, sizes.get(ROOT_ECLIPSE)).addTo(breadCrumbs);
		out.println("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">");
		out.println("<html>");
		out.println("<head>");
		out.println("  <title>");
		out.println("    " + location);
		out.println("  </title>");
		out.println("</head>");
		out.println("<body>");
		out.println("<h3>");

		for (int i = 0, limit = breadCrumbs.size() - 1; i <= limit; i++) {
			BreadCrumb breadCrumb = breadCrumbs.get(i);
			String html = "<span style='white-space: nowrap;'>" + breadCrumb.toHTML(i == limit) //
					+ (i == limit
							? " <small style='font-size: 0.6em;'>&bull; " + dates.get(ROOT_ECLIPSE) + " "
									+ times.get(ROOT_ECLIPSE) + "</small>"
							: "") //
					+ "</span>";
			out.println("    " + html);
		}

		out.println("</h3>");
		out.println("<table>");
		out.println("  <tr>");
		out.println("  <th style='text-align: left; padding-right: 2em;'>Name</th>"
				+ "<th style='padding-right: 2em;'>Size</th>" + "<th style='padding-right: 2em;'>Percent</th>"
				+ "<th style='text-align: left;'>Date</th>");

		return breadCrumbs.get(breadCrumbs.size() - 1).getMirorrLinkPrefix();
	}

	private void printHTMLSuffix(PrintStream out) {
		out.println("  </tr>");
		out.println("</table>");
		out.println("</body>");
	}

	private void handle(String location, String size, String date, String time, String fileCount, String fileSize) {
		sizes.put(location, size);
		dates.put(location, date);
		times.put(location, time);

		int index = location.lastIndexOf('/');
		if (index != -1) {
			String parent = location.substring(0, index);
			children.get(parent).add(location);
		}

		List<String> children = new ArrayList<String>();
		if (!"0".equals(fileCount)) {
			String files = location + "/" + FILES;
			sizes.put(files, fileSize);
			dates.put(files, "");
			times.put(files, "");
			fileCounts.put(files, fileCount);
			children.add(files);
		}

		this.children.put(location, children);
	}
}
